/*
 * backup.c
 *
 * Routines to perform that actual backup and restore of the device.
 *
 * Copyright (C) 2004 Daniel Thompson <rfbackup@redfelineninja.org.uk> 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "rfbackup.h"

#include <stdarg.h>

/* if these get too large it is possible we will need to request a larger stack */
#define MAX_DBNAME   32
#define MAX_FILENAME (3*MAX_DBNAME + 8)
#define MAX_PATHNAME 32
#define MAX_FULLNAME (MAX_FILENAME + MAX_PATHNAME)

#define A68K_CREATOR 'a68k'

/* StrNChr is implemented in terms of StrChr which ensures we get WChar support for free */
static Char *StrNChr(const Char *p, WChar c)
{
	Char *q;
	for (p = q = StrChr(p, c); p; q = (Char *) p, p = StrChr(p+1, c)) {}	
	return q;
}

static Err VFSFileDBInfoFromPath(
		UInt16 volRefNum, const Char *pathNameP,
		Char *nameP, UInt16 *attributesP, UInt16 *versionP, UInt32 *crDateP, UInt32 *modDateP,
		UInt32 *bckUpDateP, UInt32 *modNumP, MemHandle *appInfoHP, MemHandle *sortInfoHP,
		UInt32 *typeP, UInt32 *creatorP, UInt16 *numRecordsP)
{
	Err err;
	FileRef fileRef;
	
	err = VFSFileOpen(volRefNum, pathNameP, vfsModeRead, &fileRef);
	if (errNone == err) {
		err = VFSFileDBInfo(fileRef, nameP, attributesP, versionP, crDateP, modDateP, bckUpDateP,
				    modNumP, appInfoHP, sortInfoHP, typeP, creatorP, numRecordsP);
		(void) VFSFileClose(fileRef);
	}
	
	return err;
}

static Err VFSFileVPrintF(FileRef fileRef, const Char *fmt, va_list args)
{
	Char line[80];
	Int16 count;
	
	count = StrVPrintF(line, fmt, args);
	ErrNonFatalDisplayIf(count > sizeof(line), "internal string overflow");
	
	return VFSFileWrite(fileRef, count, line, NULL);
}

/*static*/ Err VFSFilePrintF(FileRef fileRef, const Char *fmt, ...)
{
	Err err;
	va_list args;
	
	va_start(args, fmt);
	err = VFSFileVPrintF(fileRef, fmt, args);
	va_end(args);
	return err;
}


static Err BakLogToFile(FileRef fileRef, UInt16 resID, ...)
{
	va_list args;
	MemHandle hdl;
	const Char *fmt;
	Err err;
	
	va_start(args, resID);
	hdl = DmGetResource('tSTR', resID);
	ErrNonFatalDisplayIf(0 == hdl, "Resource database incomplete/corrupt");
	fmt = (Char *) MemHandleLock(hdl);
	err = VFSFileVPrintF(fileRef, fmt, args);
	MemHandleUnlock(hdl);
	DmReleaseResource(hdl);
	va_end(args);
	
	return err;
}

static const Char *BakGetTime(void)
{
	DateTimeType dateTime;
	static Char time[timeStringLength+1];
	
	TimSecondsToDateTime(TimGetSeconds(), &dateTime);
	/* TODO: the format type should be fetched from the preferences */
	TimeToAscii(dateTime.hour, dateTime.minute, tfColonAMPM, time);
	return time;
}

static const Char *BakGetDate(void)
{
	DateTimeType dateTime;
	static Char date[longDateStrLength+1];
	
	TimSecondsToDateTime(TimGetSeconds(), &dateTime);
	/* TODO: the format type should be fetched from the preferences */
	DateToAscii(dateTime.month, dateTime.day, dateTime.year, dfDMYLong, date);
	return date;
}

static Boolean BakIsSpecialFilename(const Char *fname)
{
	/* TODO: would it be better to return false for .prc and .pdb and true for everything else */
	if (0 == StrCompareAscii(fname, "DATABASE.TMP") ||
	    0 == StrCompareAscii(fname, "BACKUP.LOG")) {
	    	return true;
	}
	
	return false;
}

static Boolean BakIsDatabaseInROM(UInt16 cardNo, LocalID dbId)
{
    Boolean result = false;
    LocalIDKind kind;
    UInt16 heapId;
    MemPtr handle;
    UInt16 heapFlags;

    kind = MemLocalIDKind(dbId);
    handle = MemLocalIDToGlobal(dbId, cardNo);
    if (kind == memIDPtr)
    {
        heapId = MemPtrHeapID(handle);
    }
    else
    {
        heapId = MemHandleHeapID(handle);
    }
    heapFlags = MemHeapFlags(heapId);
    result = ((heapFlags & 0x0001) == 0x0001);
    
    return result;
}

static Boolean BakIsDatabaseDerived(UInt16 cardNo, LocalID dbId)
{
	UInt32 creator = 0; /* ensure we are robust if DmDatabaseInfo did fail! */
	Err err;
    
	/* Check if database is a special 'a68k' file */
	err = DmDatabaseInfo(cardNo, dbId, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, &creator);
	ErrNonFatalDisplayIf(errNone != err, "DmDatabaseInfo returned an error");
        
        return A68K_CREATOR == creator;
}


static Char * BakStripIllegalCharacters(Char *s)
{
	static const Char legal[] = "abcdefghijklmnopqrstuvwxyz"
	                            "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	                            "0123456789-_ !%";
	static const Char hex[] = "0123456789abcdef";
	static Char t[MAX_FILENAME];
	
	Char *p;

	/* check every character of s for legality, convert any illegal characters in
	 * %xx form.
	 */
	for (p = t; *s != '\0'; p++, s++) {
		if (StrChr(legal, *s)) {
			*p = *s;
		} else {
			*p++ = '%';
			*p++ = hex[(*s/16) & 15];
			*p   = hex[*s & 15];
		}
	}
	*p = '\0';

	return t;
}

Err BakCleanupBackupDirectory(UInt16 volRefNum, const Char *pathNameP, BakCallback *cb, FileRef logFile)
{
	Err err;
	FileRef dirRef;
	UInt32 dirIterator;
	FileInfoType info;
	MemHandle strHandle;
	Char *str;
	Char strName[MAX_FILENAME];
	Char fileName[MAX_FILENAME];
	
	/* get the deleting string */
	strHandle = DmGetResource('tSTR', kStringDeleting);
	ErrNonFatalDisplayIf(0 == strHandle, "kStringDeleting is missing");
		
	err = VFSFileOpen(volRefNum, pathNameP, vfsModeRead, &dirRef);
	if (errNone != err) {
		return err;
	}
	
	/* initialize the info structure */
	info.nameP = fileName;
	info.nameBufLen = MAX_FILENAME;	
	
	/* iterate over every file in the backup directory */
	for (dirIterator = vfsIteratorStart, err = VFSDirEntryEnumerate(dirRef, &dirIterator, &info);
	     errNone == err && vfsIteratorStop != dirIterator;
	     err = VFSDirEntryEnumerate(dirRef, &dirIterator, &info)) {
	     	Char name[MAX_DBNAME];
	     	Char pathName[MAX_FULLNAME];
	     	Char *properName;
	     	Int16 properNameLen;

	     	/* guarantee no power off while restore is in progress */
	     	(void) EvtResetAutoOffTimer();

		if (BakIsSpecialFilename(fileName)) {
			continue;
		}

		/* generate the full path name for the .prc or .pdb file */
	     	StrCopy(pathName, pathNameP);
	     	StrNCat(pathName, "/", sizeof(pathName));
	     	StrNCat(pathName, fileName, sizeof(pathName));
	     
	        /* obtain the original name of the database */
	     	err = VFSFileDBInfoFromPath(volRefNum, pathName, name,
	     	                            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
	     	if (errNone != err) {
	     		ErrAlert(err);
	     		continue;
	     	}
	     	
	     	/* check that this name is what RFBackup would select for this database 
	     	 * and that a database of this name exists on the Palm
	     	 */
	     	properName = BakStripIllegalCharacters(name);
	     	properNameLen = StrLen(properName);
	     	if (0 != StrNCompare(properName, fileName, properNameLen) ||
	     	    '.' != fileName[properNameLen] ||
	     	    0 == DmFindDatabase(/* TODO: hardcoded cardID */ 0, name)) {
	     	   	/* update the GUI */
			str = (Char *) MemHandleLock(strHandle);
			StrPrintF(strName, str, name);
			MemHandleUnlock(strHandle);
			cb(NULL, strName, 0, 0);
			
			(void) VFSFileDelete(volRefNum, pathName);
			(void) BakLogToFile(logFile, kStringRemovedFromCard, name);
			
			/* as soon as we have deleted a file from the directory any outstanding iterators
			 * become obsolete. we will therefore have to walk the entire directory again.
			 */
			dirIterator = vfsIteratorStart;
		}
	}
	
	DmReleaseResource(strHandle);
	
	VFSFileClose(dirRef);
	
	return err;
}

static Boolean BakCompareDatabase(UInt16 volRefNum, const Char *pathNameP, UInt16 cardNo, LocalID dbID)
{
	Err err;
	
	UInt16 palmAttributes, cardAttributes;
	UInt16 palmVersion, cardVersion;
	UInt32 palmModDate, cardModDate;
	UInt32 palmModNum, cardModNum;
	UInt32 palmType, cardType;
	UInt32 palmCreator, cardCreator;
	
	err = DmDatabaseInfo(cardNo, dbID, 0, &palmAttributes, &palmVersion,
	                     0, &palmModDate, 0, &palmModNum, 0, 0, &palmType, &palmCreator);
	ErrNonFatalDisplayIf(errNone != err, "DmDatabaseInfo() failed after previously succeeding");
		
	err = VFSFileDBInfoFromPath(volRefNum, pathNameP, 0, &cardAttributes, &cardVersion, 0,
				    &cardModDate, 0, &cardModNum, 0, 0, &cardType, &cardCreator, 0);
	
	if (errNone == err &&
	    palmAttributes == cardAttributes && palmVersion == cardVersion &&
	    palmModDate == cardModDate && palmModNum == cardModNum &&
	    palmType == cardType && palmCreator == cardCreator) {
	    	return true;
	}
	
	return false;
}		

static Err BakBackupDatabase(UInt16 volRefNum, const Char *pathNameP, UInt16 cardNo, LocalID dbID)
{
	Char tmpName[MAX_FULLNAME];
	Char *fileName;
	Err err;
		
	/* generate the supplementary filenames */
	StrCopy(tmpName, pathNameP);
	fileName = StrNChr(tmpName, '/') + 1;
	ErrNonFatalDisplayIf(0 == (fileName - 1), "Badly formed pathNameP");
	/* this name must be 8.3 otherwise bugs in the PalmOS FAT driver start making odd things happen */
	StrCopy(fileName, "DATABASE.TMP");
	fileName = StrNChr(pathNameP, '/') + 1;

	/* delete obsolete temporary file (if it exists ) */
	(void) VFSFileDelete(volRefNum, tmpName);
		
	/* store the new version of the database */
	err = VFSExportDatabaseToFile (volRefNum, tmpName, cardNo, dbID);
	ErrNonFatalDisplayIf(errNone != err, "VFSExportDatabaseToFile() failed");
	if (errNone != err) {
		return err;
	}

	/* delete the original backup copy (if it exists) */
	(void) VFSFileDelete(volRefNum, pathNameP);
		
	/* rename the temporary copy */
	err = VFSFileRename (volRefNum, tmpName, fileName);
	ErrNonFatalDisplayIf(errNone != err, "VFSFileRename() failed");
	
	return err;
}

Err BakBackupToVFS(UInt16 volRefNum, const Char *pathNameP, BakCallback *cb)
{
	Err err;
	MemHandle strHandle[2];
	const Char *str[2];
	FileRef logFile;
	DmSearchStateType stateInfo;
	UInt16 cardNo;
	LocalID dbID;
	UInt16 i, numDatabases;

	/* ensure the backup path exists */
	ErrNonFatalDisplayIf(StrLen(pathNameP) >= MAX_PATHNAME, "pathNameP is too long");
	/* TODO: hard coded path */
	(void) VFSDirCreate(volRefNum, "/PALM");
	(void) VFSDirCreate(volRefNum, "/PALM/RFBackup");
	/* TODO: error checking (problems on emulator (when file exists vfsErrFileGeneric is returned) */

	/* open the log file */
	/* TODO: hard coded path */
	(void) VFSFileDelete(volRefNum, "/PALM/RFBackup/BACKUP.LOG");
	err = VFSFileOpen(volRefNum, "/PALM/RFBackup/BACKUP.LOG",
	                  vfsModeWrite | vfsModeCreate, &logFile);
	if (errNone != err) {
		ErrAlert(err);
	}
	(void) BakLogToFile(logFile, kStringBackupStartedAt, BakGetDate(), BakGetTime());
		
	/* clean up any obsolete files from the backup directory */
	strHandle[0] = DmGetResource('tSTR', kStringCleaningUp);
	ErrNonFatalDisplayIf(0 == strHandle[0], "kStringCleaningUp is missing");
	strHandle[1] = DmGetResource('tSTR', kStringScanning);
	ErrNonFatalDisplayIf(0 == strHandle[1], "kStringScanning is missing");
	str[0] = (Char *) MemHandleLock(strHandle[0]);
	str[1] = (Char *) MemHandleLock(strHandle[1]);
	cb(str[0], str[1], 0, 0);
	MemHandleUnlock(strHandle[0]);
	MemHandleUnlock(strHandle[1]);
	DmReleaseResource(strHandle[0]);
	DmReleaseResource(strHandle[1]);
	
	err = BakCleanupBackupDirectory(volRefNum, pathNameP, cb, logFile);
	if (errNone != err) {
		ErrAlert(err);
	}
	
	/* update the sub-title label */
	strHandle[0] = DmGetResource('tSTR', kStringBackingUp);
	ErrNonFatalDisplayIf(0 == strHandle[0], "kStringBackingUp is missing");
	str[0] = (Char *) MemHandleLock(strHandle[0]);
	cb(str[0], NULL, 0, 0);
	MemHandleUnlock(strHandle[0]);
	DmReleaseResource(strHandle[0]);	
	
	/* determine the number of databases in the system */
	numDatabases = 0;
	for (err = DmGetNextDatabaseByTypeCreator(true,  &stateInfo, 0, 0, false, &cardNo, &dbID);
	     errNone == err && 0 != dbID;
	     err = DmGetNextDatabaseByTypeCreator(false, &stateInfo, 0, 0, false, &cardNo, &dbID)) {
		if (!BakIsDatabaseInROM(cardNo, dbID) && !BakIsDatabaseDerived(cardNo, dbID)) {
	     		numDatabases++;
	     	}
	}
	
	/* iterate over every database in the system */
	for (i=0, err = DmGetNextDatabaseByTypeCreator(true,  &stateInfo, 0, 0, false, &cardNo, &dbID);
	     errNone == err && 0 != dbID;
	     i++, err = DmGetNextDatabaseByTypeCreator(false, &stateInfo, 0, 0, false, &cardNo, &dbID)) {
	     	Char name[MAX_DBNAME];
	     	Char pathName[MAX_FULLNAME];
	     	Char *fileName;
	     	UInt16 attributes, len;
	     	
	     	/* guarantee no power off while backup is in progress */
	     	(void) EvtResetAutoOffTimer();
	     	
	     	if (BakIsDatabaseInROM(cardNo, dbID) || BakIsDatabaseDerived(cardNo, dbID)) {
	     		continue;
	     	}
	     	
	     	/* get database attributes so we can generate the proper filename */
	     	err = DmDatabaseInfo(cardNo, dbID, name, &attributes, 0, 0, 0, 0, 0, 0, 0, 0, 0);

	     	/* update the GUI */
	     	cb(NULL, name, i, numDatabases);
	     	
	     	/* generate a complete filename for the database */
	     	fileName = BakStripIllegalCharacters(name);
	     	len = StrPrintF(pathName, "%s/%s%s", pathNameP, fileName,
	     	                0 == (attributes & dmHdrAttrResDB) ? ".pdb" : ".prc");
	     	ErrNonFatalDisplayIf(len > sizeof(pathName), "string overflow");
	     		     	
	     	/* perform the actual backup (if required) */
	     	if (!BakCompareDatabase(volRefNum, pathName, cardNo, dbID)) {
	     		err = BakBackupDatabase(volRefNum, pathName, cardNo, dbID);
	     		if (errNone != err) {
				/* TODO: need to translate errors here (esp. low battery) */
	     			ErrAlert(err);
	     			(void) BakLogToFile(logFile, kStringBackupFailed, name);
			} else {
				(void) BakLogToFile(logFile, kStringBackedUp, name);
	     		}
	     	} else {
	     		BakLogToFile(logFile, kStringUpToDate, name);
	     	}
	}
	
	/* make sure we will be able to restore! */
	(void) VFSDirCreate(volRefNum, "/PALM/Launcher");
	err = BakBackupDatabase(volRefNum, "/PALM/Launcher/RFBackup.prc", 0, DmFindDatabase(0, "RFBackup"));
	if (errNone != err) {
		ErrAlert(err);
		(void) BakLogToFile(logFile, kStringBackupFailed, "RFBackup");
	} else {
		(void) BakLogToFile(logFile, kStringInstalledRFBackup);
	}
			
	/* ... and we're done with the log file */		
	(void) BakLogToFile(logFile, kStringBackupCompletedAt, BakGetDate(), BakGetTime());
	VFSFileClose(logFile);
	
	return dmErrCantFind == err ? errNone : err;
}

Err BakDeleteDatabaseByName(const Char *dbName)
{
	Err err = errNone;
	UInt16 cardNo;
    
	for (cardNo=0; cardNo <= 1; cardNo++) {
        	LocalID dbID = DmFindDatabase(cardNo, dbName);       
	        if (0 != dbID && !BakIsDatabaseInROM(cardNo, dbID)) {
			UInt16 attributes;
	        	
			/* clear the read-only flag if applicable */
			err = DmDatabaseInfo(0, dbID, 0, &attributes, 0, 0, 0, 0, 0, 0, 0, 0, 0);
			ErrNonFatalDisplayIf(errNone != err, "cannot fetch database info");
			if (0 != (attributes & dmHdrAttrReadOnly)) {
				attributes &= ~dmHdrAttrReadOnly;
				err = DmSetDatabaseInfo(0, dbID, 0, &attributes, 0, 0, 0, 0, 0, 0, 0, 0, 0);
				ErrNonFatalDisplayIf(errNone != err, "cannot set database info");
			}
			
			/* delete the original database */
			err = DmDeleteDatabase(cardNo, dbID);
				
			/* if the DB is protected then break the protection. this is (probably)
			 * safe because RFBackup prevents the switch to another application
			 * after restore so the reset is unavoidable. happily reset breaks all
			 * the protections as well.
			 * NOTE: we don't do this in a loop since that annoys the debug ROM
			 */
			if (dmErrDatabaseProtected == err) {
				DmDatabaseProtect(0, dbID, false);
				err = DmDeleteDatabase(0, dbID);
			}
			
			break;
	        }
	}
	    
	return err;
}

Err BakRestoreFromVFS(UInt16 volRefNum, const Char *pathNameP, BakCallback *cb)
{
	Err err;
	FileRef dirRef;
	UInt32 dirIterator;
	FileInfoType info;
	static Char fileName[MAX_FILENAME];
	UInt16 i, numFiles;
	MemHandle strHandle;
	const Char *str;
		
	err = VFSFileOpen(volRefNum, pathNameP, vfsModeRead, &dirRef);
	if (errNone != err) {
		return err;
	}
	
	/* update the sub-title label */
	strHandle = DmGetResource('tSTR', kStringRestoring);
	ErrNonFatalDisplayIf(0 == strHandle, "kStringRestoring is missing");
	str = (Char *) MemHandleLock(strHandle);
	cb(str, NULL, 0, 0);
	MemHandleUnlock(strHandle);
	DmReleaseResource(strHandle);
	
	/* initialize the info structure */
	info.nameP = fileName;
	info.nameBufLen = MAX_FILENAME;
	
	/* determine the number of files in the backup directory */
	for (i=0, dirIterator = vfsIteratorStart, err = VFSDirEntryEnumerate(dirRef, &dirIterator, &info);
	     errNone == err && vfsIteratorStop != dirIterator;
	     i++, err = VFSDirEntryEnumerate(dirRef, &dirIterator, &info)) {
		/* do nothing */
	}
	numFiles = i;
	     	
	/* iterate over every file in the backup directory */
	for (i=0, dirIterator = vfsIteratorStart, err = VFSDirEntryEnumerate(dirRef, &dirIterator, &info);
	     errNone == err && vfsIteratorStop != dirIterator;
	     i++, err = VFSDirEntryEnumerate(dirRef, &dirIterator, &info)) {
	     	Char name[MAX_DBNAME];
	     	UInt32 creator;
	     	Char pathName[MAX_FULLNAME];
   		UInt16 cardNo;
		LocalID dbID;

	     	/* guarantee no power off while restore is in progress */
	     	(void) EvtResetAutoOffTimer();

		if (BakIsSpecialFilename(fileName)) {
			continue;
		}

		/* generate the full path name for the .prc or .pdb file */
	     	StrCopy(pathName, pathNameP);
	     	StrNCat(pathName, "/", sizeof(pathName));
	     	StrNCat(pathName, fileName, sizeof(pathName));
	     
	        /* obtain the original name of the database */
	     	err = VFSFileDBInfoFromPath(volRefNum, pathName, name,
	     	                            0, 0, 0, 0, 0, 0, 0, 0, 0, &creator, 0);
	     	if (errNone != err) {
	     		ErrAlert(err);
	     		continue;
	     	}
	     	
	     	/* skip restore of RFBackup itself */
	     	if (creator == CREATOR_CODE) {
	     		continue;
	     	}
	     	
	     	/* skip restore of HostFS which is just to wierd to work with (it is
	     	 * reported at protected if one tries to delete it but it cannot be
	     	 * unprotected)
	     	 */
	     	if (0 == StrCompare("HostFS", name)) {
	     		continue;
	     	}
	     	
	     	/* update the gui */
	     	cb(NULL, name, i, numFiles);
	     	
		/* delete the original database */
		err = BakDeleteDatabaseByName(name);
		if (errNone != err) {
			ErrAlert(err);
			continue;
		}
	     	
	     	/* finally restore the database from the backup */
	     	err = VFSImportDatabaseFromFile(volRefNum, pathName, &cardNo, &dbID);
	      	if (errNone != err) {
	     		break;
		}	
	} 
	
	VFSFileClose(dirRef);
	
	return err;
}