/*
 * mainform.c
 *
 * Manage the main form
 *
 * Copyright (C) 2004 Daniel Thompson <d.thompson@gmx.net> 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "rfbackup.h"

static void FldUpdateScrollBar(FormPtr form, UInt16 fieldID, UInt16 scrollbarID)
{
	FieldPtr field = (FieldPtr) FrmGetObjectPtr(form, FrmGetObjectIndex(form, fieldID));
	ScrollBarPtr scroll = (ScrollBarPtr) FrmGetObjectPtr(form, FrmGetObjectIndex(form, scrollbarID));
	UInt16 currentPosition, textHeight, fieldHeight, maxValue;
	
	FldGetScrollValues(field, &currentPosition, &textHeight, &fieldHeight);
	maxValue = (textHeight > fieldHeight ? textHeight - fieldHeight : currentPosition);
	SclSetScrollBar(scroll, currentPosition, 0, maxValue, fieldHeight - 1);
}

static void FldScrollLines(FormPtr form, UInt16 fieldID, UInt16 scrollbarID, Int16 numLines)
{
	FieldPtr field = (FieldPtr) FrmGetObjectPtr(form, FrmGetObjectIndex(form, fieldID));

	if (numLines < 0) {
		FldScrollField(field, -numLines, winUp);
		
		/* when scrolling up trailing blank lines are automagically removed, we must update
		 * the scrollbar or it will become inconsistant with the field.
		 */
		FldUpdateScrollBar(form, fieldID, scrollbarID);
	} else {
		FldScrollField(field, numLines, winDown);
	}
}

static void FldPageScroll(FormPtr form, UInt16 fieldID, UInt16 scrollbarID, WinDirectionType direction)
{
	FieldPtr field = (FieldPtr) FrmGetObjectPtr(form, FrmGetObjectIndex(form, fieldID));

	if (FldScrollable(field, direction)) {
		UInt16 numLines = FldGetVisibleLines(field) - 1;
		FldScrollField(field, numLines, direction);
		FldUpdateScrollBar(form, fieldID, scrollbarID);
	}	
}

static void FldSetTextHandleAndFree(FieldPtr field, MemHandle memHandle)
{
	MemHandle oldHandle;
	
	oldHandle = FldGetTextHandle(field);
	FldSetTextHandle(field, memHandle);
	if (oldHandle) {
		if (MemHandleDataStorage(oldHandle)) {
			DmReleaseResource(oldHandle);
		} else {
			MemHandleFree(oldHandle);
		}
	}
}
static void UpdatePreviousBackup(FormPtr form)
{
	Err err;
	UInt16 volRefNum;
	UInt32 volIterator = vfsIteratorStart;
	FileRef fileRef;
	UInt32 fileSize, actualSize;
	FieldPtr field = (FieldPtr) FrmGetObjectPtr(form, FrmGetObjectIndex(form, kFieldPreviousBackup));
	MemHandle memHandle;
	MemPtr memPtr;
		
	err = VFSVolumeEnumerate(&volRefNum, &volIterator);
	if (errNone != err) {
		FldSetTextHandleAndFree(field, DmGetResource('tSTR', kStringNoMemoryCard));
		return;
	}
	
	err = VFSFileOpen(volRefNum, "/PALM/RFBackup/BACKUP.LOG", vfsModeRead, &fileRef);
	if (errNone != err) {
		FldSetTextHandleAndFree(field, DmGetResource('tSTR', kStringNoPreviousBackup));
		return;		
	}	
	
	err = VFSFileSize(fileRef, &fileSize);
	ErrNonFatalDisplayIf(errNone != err, "Cannot obtain file size");
	
	/* now read the actual file into a memory handle */
	memHandle = MemHandleNew(fileSize + 1);
	memPtr = MemHandleLock(memHandle);
	err = VFSFileRead(fileRef, fileSize, memPtr, &actualSize);
	((Char *) memPtr)[fileSize] = '\0';
	ErrNonFatalDisplayIf(errNone != err, "Cannot read backup.log");
	ErrNonFatalDisplayIf(fileSize != actualSize, "Cannot read all of backup.log");
	MemHandleUnlock(memHandle);
	FldSetTextHandleAndFree(field, memHandle);
	
	VFSFileClose(fileRef);

	FldUpdateScrollBar(form, kFieldPreviousBackup, kScrollBarPreviousBackup);
}

static void MainFormInit(FormPtr form)
{	
	UpdatePreviousBackup(form);
}

static void MainFormDestroy(FormPtr form)
{
	FieldPtr field = (FieldPtr) FrmGetObjectPtr(form, FrmGetObjectIndex(form, kFieldPreviousBackup));
	FldSetTextHandleAndFree(field, NULL);
}

Boolean MainFormHandleEvent(EventPtr event)
{
	Boolean eventHandled = true;
	FormPtr form;	

	form = FrmGetActiveForm();
	
	switch (event->eType) {
	case frmOpenEvent:
		MainFormInit(form);
		/*FALLTHRU*/
	case frmUpdateEvent:
		FrmDrawForm(form);
		/* draw any gadgets here */
		break;
	case frmCloseEvent:
		MainFormDestroy(form);
		/* do *not* 'handle' the event - it will suppress tidying up */
		eventHandled = false;
		break;
	case ctlSelectEvent:
		switch (event->data.ctlSelect.controlID) {			
		case kButtonBackup:
			if (0 == FrmAlert(kAlertBackupConfirmation)) {
				ProgressFormDoDialog(progressFormBackupToVFS);
				UpdatePreviousBackup(form);
				FrmDrawForm(form);
			}
			break;
		case kButtonRestore:
			if (0 == FrmAlert(kAlertRestoreConfirmation)) {
				ProgressFormDoDialog(progressFormRestoreFromVFS);
			}
			break;
		default:
			ErrNonFatalDisplay("Invalid control select event");
		}
	case fldChangedEvent:
		FldUpdateScrollBar(form, kFieldPreviousBackup, kScrollBarPreviousBackup);
		break;
	case sclRepeatEvent:
		FldScrollLines(form, kFieldPreviousBackup, kScrollBarPreviousBackup,
		               event->data.sclRepeat.newValue - event->data.sclRepeat.value);
	case keyDownEvent:
		/* on 2-byte locales (e.g. Japanese) vchrPageUp/vchrPageDown collides with a real
		 * character so we have to check that is really is a virtual before we handle it.
		 */
		if (!TxtCharIsVirtual(event->data.keyDown.modifiers, event->data.keyDown.chr)) {
			eventHandled = false;
			break;
		}
		
		switch (event->data.keyDown.chr) {
		case vchrPageUp:
			FldPageScroll(form, kFieldPreviousBackup, kScrollBarPreviousBackup, winUp);
			break;
		case vchrPageDown:
			FldPageScroll(form, kFieldPreviousBackup, kScrollBarPreviousBackup, winDown);
			break;
		default:
			eventHandled = false;
		}
		break;
	default:
		eventHandled = false;
		break;
	}

	return eventHandled;
}
